@@ -7,8 +7,9 @@ from typing import List, Iterable, Tuple, Optional
 import matplotlib.pyplot as plt
 import numpy as np
 import streamlit as st
+from streamlit.delta_generator import DeltaGenerator
 
-from tnmf.TransformInvariantNMF import TransformInvariantNMF
+from tnmf.TransformInvariantNMF import TransformInvariantNMF, MiniBatchAlgorithm
 from tnmf.utils.signals import generate_pulse_train, generate_block_image
 
 HELP_CHANNEL = \
@@ -17,7 +18,12 @@ HELP_CHANNEL = \
     factorization in the sense that each atom always covers all channels.'''
 
 
-def st_define_nmf_params(default_params: dict, verbose: bool = True) -> dict:
+def explanation(text: str, verbose: bool):
+    if verbose:
+        st.sidebar.caption(text)
+
+
+def st_define_nmf_params(default_params: dict, have_ground_truth: bool = True, verbose: bool = True) -> dict:
     """
     Defines all necessary NMF parameters via streamlit widgets.
 
@@ -25,6 +31,9 @@ def st_define_nmf_params(default_params: dict, verbose: bool = True) -> dict:
     ----------
     default_params : dict
         Contains the default parameters that are used if the created streamlit checkbox is True.
+    have_ground_truth : bool
+        If True, the parameters in default_params are considered being ground truth values and
+        the respective explanatory texts are modified accordingly.
     verbose : bool
         If True, show detailed information.
 
@@ -35,54 +44,102 @@ def st_define_nmf_params(default_params: dict, verbose: bool = True) -> dict:
     """
     st.sidebar.markdown('# TNMF settings')
 
-    # decide if ground truth atom number shall be used
-    help_use_n_atoms = \
-        '''If selected, the **number of atoms** used by the model is set to the actual number of atoms used
-        for signal generation.'''
-    use_n_atoms = st.sidebar.checkbox('Use ground truth number of atoms', True, help=help_use_n_atoms)
-    if verbose:
-        st.sidebar.caption(help_use_n_atoms)
+    help_n_atoms = 'The **number of atoms**.'
+    if have_ground_truth:
+        # decide if ground truth atom number shall be used
+        help_n_atoms += ' To use the ground truth dictionary size, tick the checkbox above.'
+        help_use_n_atoms = \
+            '''If selected, the **number of atoms** used by the model is set to the actual number of atoms used
+            for signal generation.'''
+        use_n_atoms = st.sidebar.checkbox('Use ground truth number of atoms', True, help=help_use_n_atoms)
+        explanation(help_use_n_atoms, verbose)
+    else:
+        use_n_atoms = False
 
-    help_n_atoms = 'The **number of atoms**. To use the ground truth dictionary size, tick the checkbox above.'
-    n_atoms = st.sidebar.number_input('# Atoms', value=default_params['n_atoms'], min_value=1,
-                                      help=help_n_atoms) if not use_n_atoms else None
-    if verbose and not use_n_atoms:
-        st.sidebar.caption(help_n_atoms)
+    n_atoms_default = default_params['n_atoms']
+    n_atoms = st.sidebar.number_input('# Atoms', value=n_atoms_default, min_value=1,
+                                      help=help_n_atoms) if not use_n_atoms else n_atoms_default
+    explanation(help_n_atoms, verbose and not use_n_atoms)
 
-    # decide if ground truth atom shape shall be used
-    help_use_atom_shape = \
-        '''If selected, the **size of the atoms** used by the model is set the actual size of the atoms used
-        for signal generation.'''
-    use_atom_shape = st.sidebar.checkbox('Use ground truth atom size', True, help=help_use_atom_shape)
-    if verbose:
-        st.sidebar.caption(help_use_atom_shape)
+    help_atom_shape = 'The **size of each atom** dimension.'
+    if have_ground_truth:
+        # decide if ground truth atom shape shall be used
+        help_atom_shape += ' To use the ground truth atom size, tick the checkbox above.'
+        help_use_atom_shape = \
+            '''If selected, the **size of the atoms** used by the model is set the actual size of the atoms used
+            for signal generation.'''
+        use_atom_shape = st.sidebar.checkbox('Use ground truth atom size', True, help=help_use_atom_shape)
+        explanation(help_use_atom_shape, verbose)
+    else:
+        use_atom_shape = False
 
-    help_atom_shape = 'The **size of each atom** dimension. To use the ground truth atom size, tick the checkbox above.'
+    default_atom_shape = default_params['atom_shape']
     atom_shape = tuple([st.sidebar.number_input('Atom size',
-                        value=default_params['atom_shape'][0], min_value=1,
-                        help=help_atom_shape)] * len(default_params['atom_shape'])) if not use_atom_shape else None
-    if verbose and not use_atom_shape:
-        st.sidebar.caption(help_atom_shape)
+                        value=default_atom_shape[0], min_value=1,
+                        help=help_atom_shape)] * len(default_params['atom_shape'])
+                       ) if not use_atom_shape else default_atom_shape
+    explanation(help_atom_shape, verbose and not use_atom_shape)
 
     help_sparsity_H = 'The strength of the **L1 activation sparsity regularization** imposed on the optimization problem.'
     sparsity_H = st.sidebar.number_input('Activation sparsity', min_value=0.0, value=0.0, step=0.01,
                                          help=help_sparsity_H)
-    if verbose:
-        st.sidebar.caption(help_sparsity_H)
+    explanation(help_sparsity_H, verbose)
 
     help_inhibition_strength = \
-        '''The strength of the **lateral activation sparsity regularization** imposed on the optimization problem. The
-        parameter controls how strong the activation of an atom at a particular shift location suppresses the activation of
-        the same atom at neighboring locations.'''
-    inhibition_strength = st.sidebar.number_input('Lateral activation inhibition', min_value=0.0, value=0.1, step=0.01,
+        '''The strength of the **same-atom lateral activation sparsity regularization** imposed on the optimization problem.
+        The parameter controls how strong the activation of an atom at a particular shift location suppresses the activation
+        of *the same atom* at neighboring locations.'''
+    inhibition_strength = st.sidebar.number_input('Lateral activation inhibition (same atom)',
+                                                  min_value=0.0, value=0.1, step=0.01,
                                                   help=help_inhibition_strength)
-    if verbose:
-        st.sidebar.caption(help_inhibition_strength)
+    explanation(help_inhibition_strength, verbose)
 
-    help_n_iterations = '''The **number of multiplicative updates** to the atom dictionary and activation tensors.'''
-    n_iterations = st.sidebar.number_input('# Iterations', value=100, min_value=1, help=help_n_iterations)
-    if verbose:
-        st.sidebar.caption(help_n_iterations)
+    help_cross_atom_inhibition_strength = \
+        '''The strength of the **cross-atom lateral activation sparsity regularization** imposed on the optimization problem.
+        The parameter controls how strong the activation of an atom at a particular shift location suppresses the activation
+        of *all other atoms* at the same and neighboring locations.'''
+    cross_atom_inhibition_strength = st.sidebar.number_input('Lateral activation inhibition (cross-atom)',
+                                                             min_value=0.0, value=0.1, step=0.01,
+                                                             help=help_cross_atom_inhibition_strength)
+    explanation(help_cross_atom_inhibition_strength, verbose)
+
+    help_minibatch = \
+        '''Process the samples in **minibatches** instead of the full data set at once.'''
+    minibatch_updates = st.sidebar.checkbox('Minibatch updates', value=True, help=help_minibatch)
+    explanation(help_minibatch, verbose)
+    if not minibatch_updates:
+        help_n_iterations = '''The **number of multiplicative updates** to the atom dictionary and activation tensors.'''
+        n_iterations = st.sidebar.number_input('# Iterations', value=100, min_value=1, help=help_n_iterations)
+        explanation(help_n_iterations, verbose)
+    else:
+        help_algorithm = '''The **minibatch update algorithm** to be used.'''
+        algorithm = st.sidebar.radio('Minibatch algorithm', [
+            '4 - Cyclic MiniBatch for MU rules',
+            '5 - Asymmetric SG MiniBatch MU rules (ASG-MU)',
+            '6 - Greedy SG MiniBatch MU rules (GSG-MU)',
+            '7 - Asymmetric SAG MiniBatch MU rules (ASAG-MU)',
+            '8 - Greedy SAG MiniBatch MU rules (GSAG-MU)'],
+                                     1, help=help_algorithm)
+        algorithm = MiniBatchAlgorithm(int(algorithm[0]))
+        explanation(help_algorithm, verbose)
+
+        help_epoch = '''The number of **passes through the whole data set**.'''
+        n_epochs = st.sidebar.number_input('# Epochs', value=100, min_value=1, help=help_epoch)
+        explanation(help_epoch, verbose)
+
+        help_batch_size = '''The number of **samples per batch**.'''
+        batch_size = st.sidebar.number_input('# Batch size', value=3, min_value=1, help=help_batch_size)
+        explanation(help_batch_size, verbose)
+
+        sag_lambda = None
+        if algorithm in (MiniBatchAlgorithm.ASAG_MU, MiniBatchAlgorithm.GSAG_MU):
+            help_sag_lambda = \
+                '''The **exponential forgetting factor** for for the stochastic **average** gradient updates. A value of 1.0
+                means that only the latest minibatch is used for the update. The smaller the value, the more weight is put on
+                older minibatches.'''
+            sag_lambda = st.sidebar.number_input('Lambda', min_value=0.0, max_value=1., value=0.2, step=0.01,
+                                                 help=help_sag_lambda)
+            explanation(help_sag_lambda, verbose)
 
     help_backend = \
         '''The **optimization backend** for computing the multiplicative gradients.
@@ -90,8 +147,7 @@ def st_define_nmf_params(default_params: dict, verbose: bool = True) -> dict:
             for speed comparisons only.'''
     backend = st.sidebar.selectbox('Backend', ['numpy', 'numpy_fft', 'numpy_caching_fft', 'pytorch', 'pytorch_fft'], 4,
                                    help=help_backend)
-    if verbose:
-        st.sidebar.caption(help_backend)
+    explanation(help_backend, verbose)
 
     help_reconstruction_mode = \
         '''Defines the **convolution mode** for the signal reconstruction.\
@@ -109,29 +165,34 @@ def st_define_nmf_params(default_params: dict, verbose: bool = True) -> dict:
         input array are inserted circularly on the respective other side of the array.'''
     reconstruction_mode = st.sidebar.selectbox('Reconstruction', ['valid', 'full', 'circular'], 2,
                                                help=help_reconstruction_mode)
-    if verbose:
-        st.sidebar.caption(help_reconstruction_mode)
+    explanation(help_reconstruction_mode, verbose)
 
-    selected_params = dict(
-        n_iterations=n_iterations,
+    nmf_params = dict(
         n_atoms=n_atoms,
         atom_shape=atom_shape,
-        sparsity_H=sparsity_H,
-        inhibition_strength=inhibition_strength,
         backend=backend,
         reconstruction_mode=reconstruction_mode,
     )
 
-    # default parameter selection dictionary
-    use_default = dict(
-        n_atoms=use_n_atoms,
-        atom_shape=use_atom_shape,
-    )
+    if not minibatch_updates:
+        fit_params = dict(
+            n_iterations=n_iterations,
+            sparsity_H=sparsity_H,
+            inhibition_strength=inhibition_strength,
+            cross_atom_inhibition_strength=cross_atom_inhibition_strength,
+        )
+    else:
+        fit_params = dict(
+            algorithm=algorithm,
+            n_epochs=n_epochs,
+            batch_size=batch_size,
+            sag_lambda=sag_lambda,
+            sparsity_H=sparsity_H,
+            inhibition_strength=inhibition_strength,
+            cross_atom_inhibition_strength=cross_atom_inhibition_strength,
+        )
 
-    # overwrite selected parameters with defaults
-    selected_params.update({k: v for k, v in default_params.items() if use_default[k]})
-
-    return selected_params
+    return nmf_params, fit_params
 
 
 class SignalTool(ABC):
@@ -178,8 +239,7 @@ class SignalTool(ABC):
             '''The **number of generated signals** passed as input to the algorithm.
             All signals have the same shape and number of channels.'''
         n_signals = st.sidebar.number_input('# Signals', min_value=1, value=10, help=help_n_signals)
-        if verbose:
-            st.sidebar.caption(help_n_signals)
+        explanation(help_n_signals, verbose)
         signal_params = cls.st_define_signal_params(verbose=verbose)
 
         # create the input
@@ -224,7 +284,7 @@ class SignalTool(ABC):
             ''')
 
         # show the input, its reconstruction, and the reconstruction error as images next to each other
-        cols = st.beta_columns(3)
+        cols = st.columns(3)
         for col, X, title in zip(cols, [V, R, V-R], ['Input', 'Reconstruction', 'Error']):
             with col:
                 fig = plt.figure()
@@ -284,7 +344,7 @@ class SignalTool(ABC):
         i_signal = st.slider('Signal number', 1, V.shape[0], key='i_signal_partial') - 1 if V.shape[0] > 1 else 0
         for i_atom in range(nmf.n_atoms):
             R_atom = nmf.R_partial(i_atom)
-            cols = st.beta_columns(2)
+            cols = st.columns(2)
             for col, signals, signal_opts, opts in zip(
                     cols,
                     [[nmf.W[i_atom]], [R_atom[i_signal], V[i_signal]]],
@@ -368,16 +428,14 @@ class SignalTool1D(SignalTool):
 
         # define the number of channels
         n_channels = st.sidebar.number_input('# Channels', min_value=1, value=3, max_value=5, help=HELP_CHANNEL)
-        if verbose:
-            st.sidebar.caption(HELP_CHANNEL)
+        explanation(HELP_CHANNEL, verbose)
 
         # define the pulse shapes
         help_shapes = \
             '''The **pulse shapes** that can be combined along different channels to form a multi-channel time-series
             symbol.'''
         shapes = st.sidebar.multiselect('Pulse shapes', ['n', '-', '^', 'v', '_'], ['n', '-', '^', 'v', '_'], help=help_shapes)
-        if verbose:
-            st.sidebar.caption(help_shapes)
+        explanation(help_shapes, verbose)
 
         # to avoid having too many different symbols, consider only those where all channels are identical
         help_symbols = \
@@ -385,20 +443,17 @@ class SignalTool1D(SignalTool):
              signal generation. The subset of channel-symmetric symbols is pre-selected.'''
         symbols = st.sidebar.multiselect('Symbols', [''.join(chars) for chars in product(*repeat(shapes, n_channels))],
                                          [s * n_channels for s in shapes], help=help_symbols)
-        if verbose:
-            st.sidebar.caption(help_symbols)
+        explanation(help_symbols, verbose)
 
         # define the number of pulses
         help_n_pulses = 'The **number of symbols** concatenated randomly to form a signal.'
         n_pulses = st.sidebar.number_input('# Symbols', min_value=1, value=3, help=help_n_pulses)
-        if verbose:
-            st.sidebar.caption(help_n_pulses)
+        explanation(help_n_pulses, verbose)
 
         # define the pulse length
         help_pulse_length = 'The **length** of each individual multi-channel symbol.'
         pulse_length = st.sidebar.number_input('Symbol length', min_value=1, value=20, help=help_pulse_length)
-        if verbose:
-            st.sidebar.caption(help_pulse_length)
+        explanation(help_pulse_length, verbose)
 
         # create the parameter dictionary
         signal_params = dict(
@@ -444,8 +499,7 @@ class SignalTool2D(SignalTool):
         # choose between grayscale or color images
         n_channels = st.sidebar.radio('# Channels', ['1 (Grayscale images)', '3 (Color images)'], 0, help=HELP_CHANNEL)
         n_channels = 1 if n_channels == '1 (Grayscale images)' else 3
-        if verbose:
-            st.sidebar.caption(HELP_CHANNEL)
+        explanation(HELP_CHANNEL, verbose)
 
         # create all possible combinations of shapes and color
         shapes = ['+', 'x', 's']
@@ -459,22 +513,19 @@ class SignalTool2D(SignalTool):
             images) defines the color of the symbol. For color images, a representative subset of symbols is pre-selected.'''
         symbols = st.sidebar.multiselect('Symbols', all_symbols, ['+r', 'xg', 'sb'] if n_channels == 3 else all_symbols,
                                          help=help_symbols)
-        if verbose:
-            st.sidebar.caption(help_symbols)
+        explanation(help_symbols, verbose)
 
         # select the number of symbols per dimension
         help_n_symbols = \
             '''The **number of symbols** placed randomly next to each other both horizontally and vertically to generate an
             input image.'''
         n_symbols = st.sidebar.number_input('# Symbols per dimension', min_value=1, value=5, help=help_n_symbols)
-        if verbose:
-            st.sidebar.caption(help_n_symbols)
+        explanation(help_n_symbols, verbose)
 
         # define the size of each symbol
         help_symbol_size = 'The **size of each symbol** in pixels per dimension. All symbols are of square size.'
         symbol_size = st.sidebar.number_input('Symbol size', min_value=1, value=10, help=help_symbol_size)
-        if verbose:
-            st.sidebar.caption(help_symbol_size)
+        explanation(help_symbol_size, verbose)
 
         # create the parameter dictionary
         signal_params = dict(
@@ -492,7 +543,7 @@ class SignalTool2D(SignalTool):
 
     @classmethod
     def _st_compare_individual_signals(cls, V_i: np.ndarray, R_i: np.ndarray):
-        cols = st.beta_columns(2)
+        cols = st.columns(2)
         for col, X, title in zip(cols, [V_i, R_i], ['Input', 'Reconstruction']):
             with col:
                 cls.plot_signals([X], opts={'title': title})
@@ -503,3 +554,13 @@ class SignalTool2D(SignalTool):
         plt.imshow(signals[0].transpose((1, 2, 0)) / signals[0].max())
         plt.title(opts.get('title'))
         st.pyplot(fig)
+
+
+# TODO: replace st.cache with st.memo + remove deepcopies when called
+# currently this is not possible because the experimental memo function fails at hashing MiniBatchAlgorithm Enum instances
+@st.cache(hash_funcs={DeltaGenerator: lambda _: None}, allow_output_mutation=True)
+def fit_nmf_model(V, nmf_params, fit_params, progress_bar):
+    nmf = TransformInvariantNMF(**nmf_params)
+    n_steps = fit_params['n_iterations'] if 'n_iterations' in fit_params else fit_params['n_epochs']
+    nmf.fit(V, progress_callback=lambda _, x: progress_bar.progress((x + 1) / n_steps), **fit_params)
+    return nmf
